﻿//
// Copyright (C) 2009 Chris Dziemborowicz
//
// This file is part of Orzeszek Ratings.
//
// Orzeszek Ratings is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// Orzeszek Ratings is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Interop;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using iTunesLib;
using WMPLib;

namespace OrzeszekRatings
{
	/// <summary>
	/// Interaction logic for SyncProgressControl.xaml
	/// </summary>
	public partial class SyncProgressControl : UserControl
	{
		private enum SyncState
		{
			Initializing,
			Syncing,
			Done,
			Cancelled,
			Error
		}

		private object syncLock = new object();
		private Thread thread;

		private SyncState state;
		public double progress;
		private string currentFile;
		private string currentError;

		private bool cancelled;

		public SyncInfo SyncInfo { get; private set; }
		public WindowInteropHelper InteropHelper { get; set; }

		public event EventHandler Done;
		public event EventHandler CloseRequested;
        System.IO.StreamWriter file = new System.IO.StreamWriter(System.AppDomain.CurrentDomain.BaseDirectory + "\\SyncLog.txt", true);

		public SyncProgressControl()
		{
			InitializeComponent();
        }

		public void AnimateShow()
		{
			Visibility = Visibility.Visible;

			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(0, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			da.KeyFrames.Add(new SplineDoubleKeyFrame(1, KeyTime.FromTimeSpan(TimeSpan.FromMilliseconds(250))));
			MainSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}

		public void AnimateHide()
		{
			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(1, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			da.KeyFrames.Add(new SplineDoubleKeyFrame(0, KeyTime.FromTimeSpan(TimeSpan.FromMilliseconds(250))));
			da.Completed += new EventHandler(delegate(object o, EventArgs ea)
				{
					Visibility = Visibility.Collapsed;
				}
			);
			MainSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}

        private void Logger (string logtext)
        {
            file.WriteLine(logtext);
            file.Flush();
        }

        private void Logger (string trackName, string todirection, string property, string formerValue, string newValue)
        {
            todirection = todirection=="toWMP"?"iTunes to WMP":"WMP to iTunes";
            string logtext = "File: " + trackName + ";Direction: " + todirection + "; Property: " + property + "; Former Value: " + formerValue + "; newvalue: " + newValue;
            file.WriteLine(logtext);
            file.Flush();
        }

        public void StartSync(SyncInfo syncInfo)
		{
			if (SyncInfo != null)
				throw new InvalidOperationException();

			if (syncInfo.Direction == SyncDirection.None || (!syncInfo.SyncPlayCounts && !syncInfo.SyncRatings))
				throw new ArgumentException();

			SyncInfo = syncInfo;

			if (syncInfo.Direction == SyncDirection.FromWmp)
				if (syncInfo.SyncPlayCounts && syncInfo.SyncRatings)
					MainStatusTextBlock.Text = "Syncing play counts and ratings from WMP to iTunes";
				else if (syncInfo.SyncPlayCounts)
					MainStatusTextBlock.Text = "Syncing play counts from WMP to iTunes";
				else
					MainStatusTextBlock.Text = "Syncing ratings from WMP to iTunes";
			else if (syncInfo.Direction == SyncDirection.ToWmp)
				if (syncInfo.SyncPlayCounts && syncInfo.SyncRatings)
					MainStatusTextBlock.Text = "Syncing play counts and ratings from iTunes to WMP";
				else if (syncInfo.SyncPlayCounts)
					MainStatusTextBlock.Text = "Syncing play counts from iTunes to WMP";
				else
					MainStatusTextBlock.Text = "Syncing ratings from iTunes to WMP";
            else
                if (syncInfo.SyncPlayCounts && syncInfo.SyncRatings)
                    MainStatusTextBlock.Text = "Syncing play counts and ratings between iTunes and WMP";
                else if (syncInfo.SyncPlayCounts)
                    MainStatusTextBlock.Text = "Syncing play counts between iTunes and WMP";
                else
                    MainStatusTextBlock.Text = "Syncing ratings between iTunes and WMP";

			state = SyncState.Initializing;
			progress = 0;
			currentFile = string.Empty;
			currentError = null;
			cancelled = false;

			CancelButton.IsEnabled = true;
			OKButton.IsEnabled = false;
			CloseButton.IsEnabled = false;

			ShowCancelNow();
			HideButtonsNow();

			UpdateInterface();
			AnimateShow();

			thread = new Thread(new ThreadStart(DoWork));
			thread.Start();
		}

		public void UpdateInterface()
		{
			lock (syncLock)
				if (SyncInfo != null)
				{
					ProgressBar.Value = progress;

					if (InteropHelper != null)
						if (SyncInfo == null || progress == 100)
							TaskbarUtility.SetProgressState(InteropHelper.Handle, TaskbarProgressState.NoProgress);
						else
							TaskbarUtility.SetProgressValue(InteropHelper.Handle, (ulong)progress, 100);

					switch (state)
					{
						case SyncState.Initializing:
							SubStatusTextBlock.Text = "Initializing...";
							break;

						case SyncState.Syncing:
							SubStatusTextBlock.Text = string.IsNullOrEmpty(currentFile) ? "Initializing..." : string.Format("Syncing metadata...{0}{0}{1}", Environment.NewLine, currentFile);
							break;

						case SyncState.Done:
							SubStatusTextBlock.Text = "Done.";
							SyncInfo = null;
							break;

						case SyncState.Cancelled:
							SubStatusTextBlock.Text = "Cancelled.";
							SyncInfo = null;
							break;

						case SyncState.Error:
							SubStatusTextBlock.Text = currentError ?? "Error.";
							SyncInfo = null;
							break;
					}
				}
		}

		public void Close()
		{
			if (thread != null)
				thread.Abort();
            file.Close();
		}

		private void DoWork()
		{
			iTunesApp itunes = null;
			WindowsMediaPlayer wmp = null;

			try
			{
				itunes = new iTunesApp();
				wmp = new WindowsMediaPlayer();

				int wmpPlayCountAtom = wmp.mediaCollection.getMediaAtom("UserPlayCount");
				int wmpRatingAtom = wmp.mediaCollection.getMediaAtom("UserRating");
                int wmpLastPlayedTimeAtom = wmp.mediaCollection.getMediaAtom("UserLastPlayedTime");
				state = SyncState.Syncing;
                Logger("Starting sync at " + DateTime.Now.ToString());
				IITTrackCollection itunesTracks = itunes.LibraryPlaylist.Tracks;
                int nullFileCount=0;

				for (int i = 1; i <= itunesTracks.Count; i++)
				{
					if (cancelled)
					{
						state = SyncState.Cancelled;

						if (wmp != null)
							wmp.close();

						return;
					}

					IITTrack itunesTrack = itunesTracks[i];
					if (itunesTrack.Kind == ITTrackKind.ITTrackKindFile)
					{
						IITFileOrCDTrack itunesFileTrack = (IITFileOrCDTrack)itunesTrack;
						IWMPPlaylist wmpTrackSearch = null;

						currentFile = itunesFileTrack.Location;
                        try
                        {
                            if (itunesFileTrack.Location != null)
                            {
                                wmpTrackSearch = wmp.mediaCollection.getByAttribute("SourceURL", itunesFileTrack.Location);
                                if (wmpTrackSearch.count > 0)
                                {
                                    IWMPMedia wmpTrack = wmpTrackSearch.get_Item(0);

                                    int itunesPlayCount = itunesFileTrack.PlayedCount;
                                    int wmpPlayCount = 0;
                                    int.TryParse(wmpTrack.getItemInfoByAtom(wmpPlayCountAtom), out wmpPlayCount);

                                    int? itunesRating = itunesFileTrack.ratingKind == ITRatingKind.ITRatingKindUser ? RatingUtility.FromItunesRating(itunesFileTrack.Rating) : null;
                                    int? wmpRating = RatingUtility.FromWmpRating(wmpTrack.getItemInfoByAtom(wmpRatingAtom));

                                    if (SyncInfo.Direction == SyncDirection.FromWmp || SyncInfo.Direction == SyncDirection.BiDir)
                                    {
                                        if (SyncInfo.SyncPlayCounts)
                                            if (SyncInfo.OnlySyncHighestPlayCount)
                                                try
                                                {
                                                    if (itunesPlayCount < wmpPlayCount)
                                                    {
                                                        Logger(itunesFileTrack.Location,"toiTunes","PlayCount",itunesFileTrack.PlayedCount.ToString(), wmpPlayCount.ToString());
                                                        itunesFileTrack.PlayedCount = wmpPlayCount;
                                                    }
                                                }
                                                catch (Exception ex)
                                                {
                                                    Logger(ex.Message);
                                                }
                                            else if (SyncInfo.AggregateMovePlayCount)
                                                try
                                                {
                                                    if (wmpPlayCount != 0)
                                                    {
                                                        itunesFileTrack.PlayedCount = itunesPlayCount + wmpPlayCount;
                                                        wmpTrack.setItemInfo("UserPlayCount", "0");
                                                        Logger(itunesFileTrack.Location, "toiTunes", "Aggregating playcount", wmpPlayCount.ToString(), (itunesPlayCount + wmpPlayCount).ToString());
                                                    }
                                                }
                                                catch (Exception ex)
                                                {
                                                    Logger(ex.Message);
                                                }
                                            else
                                                try
                                                {
                                                    if (!SyncInfo.OnlySyncRatingsPlayCount || wmpPlayCount > itunesPlayCount)
                                                        if (itunesPlayCount != wmpPlayCount)
                                                        {
                                                            Logger(itunesFileTrack.Location, "toiTunes", "PlayCount", itunesFileTrack.PlayedCount.ToString(), wmpPlayCount.ToString());
                                                            itunesFileTrack.PlayedCount = wmpPlayCount;
                                                        }
                                                }
                                                catch (Exception ex)
                                                {
                                                    Logger(ex.Message);
                                                }

                                        if (SyncInfo.SyncRatings)
                                            try
                                            {
                                                if (!SyncInfo.OnlySyncRatingsPlayCount || wmpPlayCount > itunesPlayCount)
                                                    if (!SyncInfo.OnlySyncMissingRatings || itunesRating == null)
                                                        if (wmpRating != itunesRating)
                                                        {
                                                            Logger(itunesFileTrack.Location, "toiTunes", "Rating", itunesRating.ToString(), wmpRating.ToString());
                                                            itunesFileTrack.Rating = RatingUtility.ToItunesRating(wmpRating);
                                                        }
                                            }
                                            catch (Exception ex)
                                            {
                                                Logger(ex.Message);
                                            }
                                        if (SyncInfo.SyncLastPlayedDate)
                                            try
                                            {
                                                {
                                                    string wmpLastPlayedString = wmpTrack.getItemInfoByAtom(wmpLastPlayedTimeAtom);
                                                    if (wmpLastPlayedString != "")
                                                    {
                                                        DateTime wmpLastPlayed = Convert.ToDateTime(wmpLastPlayedString);
                                                        if (!SyncInfo.OnlySyncLastPlayedIfMoreRecent || wmpLastPlayed > itunesFileTrack.PlayedDate.ToUniversalTime())
                                                        {
                                                            Logger(itunesFileTrack.Location,"toiTunes","PlayedDate", itunesFileTrack.PlayedDate.ToString() ,wmpLastPlayed.ToLocalTime().ToString());
                                                            itunesFileTrack.PlayedDate = wmpLastPlayed;
                                                        }
                                                    }
                                                }
                                            }
                                            catch (Exception ex)
                                            {
                                                Logger(ex.Message);
                                            }
                                    }
                                    if (SyncInfo.Direction == SyncDirection.ToWmp || SyncInfo.Direction == SyncDirection.BiDir)
                                    {
                                        if (SyncInfo.SyncPlayCounts)
                                            if (SyncInfo.OnlySyncHighestPlayCount)
                                                try
                                                {
                                                    if (wmpPlayCount < itunesPlayCount)
                                                    {
                                                        Logger(itunesFileTrack.Location,"toWMP","PlayCount",wmpPlayCount.ToString(),itunesPlayCount.ToString());
                                                        wmpTrack.setItemInfo("UserPlayCount", itunesPlayCount.ToString());
                                                    }
                                                }
                                                catch (Exception ex)
                                                {
                                                    Logger(ex.Message);
                                                }
                                            else if (SyncInfo.AggregateMovePlayCount)
                                                try
                                                {
                                                    if (itunesPlayCount != 0)
                                                    {
                                                        Logger(itunesFileTrack.Location,"toiWMP","Aggregating playcount",itunesPlayCount.ToString(), (itunesPlayCount + wmpPlayCount).ToString());
                                                        wmpTrack.setItemInfo("UserPlayCount", (itunesPlayCount + wmpPlayCount).ToString());
                                                        itunesFileTrack.PlayedCount = 0;
                                                    }
                                                }
                                                catch (Exception ex)
                                                {
                                                    Logger(ex.Message);
                                                }
                                            else
                                                try
                                                {
                                                    if (!SyncInfo.OnlySyncRatingsPlayCount || itunesPlayCount > wmpPlayCount)
                                                        if (itunesPlayCount != wmpPlayCount)
                                                        {
                                                            Logger(itunesFileTrack.Location, "toWMP", "UserPlayCount", wmpPlayCount.ToString(), itunesPlayCount.ToString());
                                                            wmpTrack.setItemInfo("UserPlayCount", itunesPlayCount.ToString());
                                                        }
                                                }
                                                catch (Exception ex)
                                                {
                                                    Logger(ex.Message);
                                                }

                                        if (SyncInfo.SyncRatings)
                                            try
                                            {
                                                if (!SyncInfo.OnlySyncRatingsPlayCount || itunesPlayCount > wmpPlayCount)
                                                    if (!SyncInfo.OnlySyncMissingRatings || wmpRating == null)
                                                        if (wmpRating != itunesRating)
                                                        {
                                                            Logger(itunesFileTrack.Location, "toWMP", "Rating", wmpRating.ToString(), itunesRating.ToString());
                                                            wmpTrack.setItemInfo("UserRating", RatingUtility.ToWmpRating(itunesRating));
                                                        }
                                            }
                                            catch (Exception ex)
                                            {
                                                Logger(ex.Message);
                                            }
                                    }
                                }
                            }
                            else nullFileCount++;
                        }
                        catch (Exception ex)
                        {
                            Logger(ex.Message);
                        }
						finally
						{
							if (wmpTrackSearch != null)
								wmpTrackSearch.clear();
						}
					}

					progress = 100.0 * i / itunesTracks.Count;
				}

				progress = 100.0;

				if (!cancelled)
				{
					state = SyncState.Done;
					OnDone();
					StatusBarManager.AddMessage("Done.", MessageType.Success, MessageLength.Short);
				}
                //MessageBox.Show("Number of nullfiles: " + nullFileCount.ToString());
			}
			catch (Exception)
			{
				if (!cancelled)
				{
					currentError = string.Format("An unexpected error occurred.{0}{0}Make sure that WMP and iTunes are installed, and that you have the required permissions to read and modify the libraries.{0}{0}Do not close iTunes while Orzeszek Ratings is running.", Environment.NewLine);
					state = SyncState.Error;
					OnDone();
					StatusBarManager.AddMessage("An unexpected error occurred.", MessageType.Error, MessageLength.Short);
				}
			}
			finally
			{
				if (wmp != null)
					wmp.close();
			}
		}

		private void OnDone()
		{
			Dispatcher.BeginInvoke(new Action(delegate()
				{
					CancelButton.IsEnabled = false;
					OKButton.IsEnabled = true;
					CloseButton.IsEnabled = true;

					UpdateInterface();
					AnimateHideCancel();
					AnimateShowButtons();
				}
			));
		}

		private void CancelButton_Click(object sender, RoutedEventArgs e)
		{
			CancelButton.IsEnabled = false;
			OKButton.IsEnabled = true;
			CloseButton.IsEnabled = true;

			state = SyncState.Cancelled;
			cancelled = true;

			UpdateInterface();
			AnimateHideCancel();
			AnimateShowButtons();
		}

		private void OKButton_Click(object sender, RoutedEventArgs e)
		{
			lock (syncLock)
				SyncInfo = null;

			StatusBarManager.ClearIndefiniteMessages();

			if (Done != null)
				Done(this, new EventArgs());
		}

		private void CloseButton_Click(object sender, RoutedEventArgs e)
		{
			lock (syncLock)
				SyncInfo = null;

			StatusBarManager.ClearIndefiniteMessages();

			if (CloseRequested != null)
				CloseRequested(this, new EventArgs());
		}

		private void AnimateShowButtons()
		{
			ButtonsSizer.Visibility = Visibility.Visible;

			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(0, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			da.KeyFrames.Add(new SplineDoubleKeyFrame(1, KeyTime.FromTimeSpan(TimeSpan.FromMilliseconds(250))));
			ButtonsSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}

		private void AnimateHideButtons()
		{
			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(1, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			da.KeyFrames.Add(new SplineDoubleKeyFrame(0, KeyTime.FromTimeSpan(TimeSpan.FromMilliseconds(250))));
			da.Completed += new EventHandler(delegate(object o, EventArgs ea)
				{
					ButtonsSizer.Visibility = Visibility.Collapsed;
				}
			);
			ButtonsSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}

		private void AnimateShowCancel()
		{
			CancelSizer.Visibility = Visibility.Visible;

			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(0, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			da.KeyFrames.Add(new SplineDoubleKeyFrame(1, KeyTime.FromTimeSpan(TimeSpan.FromMilliseconds(250))));
			CancelSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}

		private void AnimateHideCancel()
		{
			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(1, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			da.KeyFrames.Add(new SplineDoubleKeyFrame(0, KeyTime.FromTimeSpan(TimeSpan.FromMilliseconds(250))));
			da.Completed += new EventHandler(delegate(object o, EventArgs ea)
				{
					CancelSizer.Visibility = Visibility.Collapsed;
				}
			);
			CancelSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}

		private void ShowCancelNow()
		{
			CancelSizer.Visibility = Visibility.Visible;

			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(1, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			CancelSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}

		private void HideButtonsNow()
		{
			DoubleAnimationUsingKeyFrames da = new DoubleAnimationUsingKeyFrames();
			da.KeyFrames.Add(new SplineDoubleKeyFrame(0, KeyTime.FromTimeSpan(TimeSpan.Zero)));
			da.Completed += new EventHandler(delegate(object o, EventArgs ea)
				{
					ButtonsSizer.Visibility = Visibility.Collapsed;
				}
			);
			ButtonsSizer.BeginAnimation(SizerControl.HeightFactorProperty, da);
		}
	}
}